#include "SamFile.h"
#include "Pileup.h"
#include "PileupElement.h"
#include "unistd.h"
#include "RangeList.h"
#include "Parameters.h"
#include "Error.h"
#include <map>
#include <algorithm>

class CmdParameters
{
 public:
	int min_baseq; //filter baseq. this does not affect the statistics of low mapq 
	int  min_mapq; //filter mapq. this does not affect the statistics of low baseq

        int indel_winsize;
        int indel_cnt;
        double indel_pct;
	int headClip_cnt;
	double headClip_pct;
        int mut_median_cycles2ends;
        double SB_OR;
        int zero_mapq;
	int mut_base_cnt;
	double mut_base_pct;
        int improper_paired_cnt;
        double improper_paired_pct;
        int mut_improper_paired_cnt;
        double mut_improper_paired_pct;
        double low_mapq_cutoff;
        int low_mapq_cnt;
        double low_mapq_pct;
        int mut_low_mapq_cnt;
        double mut_low_mapq_pct;
        double low_baseq_cutoff;
        double low_baseq_cnt;
        int low_baseq_pct;
        int mut_low_baseq_cnt;
        double mut_low_baseq_pct;
        double strand_bias_OR;
        int total_other_allele_cnt;
	double other_allele_pct;
	std::string outFile;

 public:
	CmdParameters(){}
	~CmdParameters(){}
};

class BedInfo
{
 public:
	std::string chr;
	int	start;
	int	end;
	char	ref;
	char	mutant;
};

typedef std::map< char, int> freqType;

template <class T, class U>
class SiteCoverage: public PileupElement {
public:
    SiteCoverage():depth(0) {
	init();
 };
    SiteCoverage(bool sum) : depth(0){
	init();
}
    virtual ~SiteCoverage() {};

    void init(){
    parset = false;
    total_ref_cnt = 0;
    total_mutant_cnt =0;
    total_other_allele_cnt =0;
    insertion_cnt =0;
    deletion_cnt = 0;
    headClip_cnt = 0;
    mut_median_dist2ends = 0;
    low_mapq_cnt = 0;
    low_baseq_cnt = 0;
    mut_low_mapq_cnt = 0;
    mut_low_baseq_cnt = 0;
    improper_paired_cnt = 0;
    mut_improper_paired_cnt = 0;
    low_mapq_pct = 0.0;;
    low_baseq_pct = 0.0;
    mut_low_mapq_pct = 0.0;
    mut_low_baseq_pct = 0.0;
    improper_paired_pct = 0.0;
    mut_improper_paired_pct = 0.0;
    mut_plus_cnt = 0;
    ref_plus_cnt = 0;
    mut_minus_cnt = 0;
    ref_minus_cnt = 0;

    for(int i=0; i<6; i++) { 
	nt_count[i] = 0; ps_nt_count[i] = 0; ns_nt_count[i] = 0; ps_nt_count_all[i] = 0; ns_nt_count_all[i] = 0; 
	nt_qual[i] = 0; ps_nt_qual[i] = 0; ns_nt_qual[i] = 0; 
	nt_mapq[i] = 0; ps_nt_mapq[i] = 0; ns_nt_mapq[i] = 0; 
	nt_cycle[i] = 0; ps_nt_cycle[i] = 0; ns_nt_cycle[i] = 0; 
        r1f=r1r=r2f=r2r=0;
    }
}

    // Add an entry to this pileup element.
    virtual void addEntry(SamRecord& record) {
        // Call the base class:
        PileupElement::addEntry(record);

        int mapq = record.getMapQuality();
	this->alldepth++;

        // Here is your code:
        Cigar* cigar = record.getCigarInfo();

	int query_map_pos = record.get0BasedPosition();
	int query_map_pos_end = record.get0BasedAlignmentEnd();

        int readIndex = cigar->getQueryIndex(query_map_pos, record.get0BasedPosition());

	int32_t left_clip_len = -record.get0BasedUnclippedStart() + record.get0BasedPosition();
	int32_t right_clip_len = record.get0BasedUnclippedEnd() - record.get0BasedAlignmentEnd();

	int pre = -1; bool insertion_flag = false; bool deletion_flag = false;
	for(int i=record.get0BasedPosition(); i<record.get0BasedAlignmentEnd(); i++)
	{
	  if(i+par.indel_winsize/2 < getRefPosition() ) continue;
	  if(i-par.indel_winsize/2 > getRefPosition() ) break;
	  readIndex = cigar->getQueryIndex(i, record.get0BasedPosition());
	  if(pre==-1) pre = readIndex-1;
	  if(deletion_flag==false && readIndex == CigarRoller::INDEX_NA)  // deletion
	   {
	  	this->deletion_cnt++;
		deletion_flag=true;
	   }
	  if(insertion_flag==false && readIndex - pre > 1) //insertion
	   {
		this->insertion_cnt++;
		insertion_flag = true;
	   }
	  pre = readIndex;
	}

        int readIndex_start = cigar->getQueryIndex(query_map_pos, record.get0BasedPosition());
	int readIndex_end = cigar->getQueryIndex(query_map_pos_end, query_map_pos);

        readIndex =  cigar->getQueryIndex(getRefPosition(), record.get0BasedPosition());

	int dist_to_left = readIndex + 1 - readIndex_start;
	int dist_to_right = readIndex_end - readIndex + 1;
	int min_dist_to_either_end = dist_to_left;
	if(min_dist_to_either_end > dist_to_right) min_dist_to_either_end = dist_to_right;

        if(readIndex != CigarRoller::INDEX_NA)
        {
            char base = record.getSequence(readIndex);
            int qual = record.getQuality(readIndex);
            this->depth++;
            this->freq[base] ++;
            this->mapq.push_back(mapq);
            this->bases.push_back(base);
            this->cycles.push_back(min_dist_to_either_end);
            this->qualities.push_back(qual);
	    this->proper.push_back( (record.getFlag() & 0x2 ) ? true : false );
	    this->headClip.push_back(record.getFlag() & 0x10 ? left_clip_len : right_clip_len);

        if (record.getFlag() & 0x10 ) { // check read strand
		  this->numMinusStrand++;
        	  this->strands.push_back('-');
		  if(record.getFlag() & 0x40) r1r++;
		  else r2r++;
        } else {
	    this->numPlusStrand++;
            this->strands.push_back('+');
	    if(record.getFlag() & 0x40) r1f++;
  	    else r2f++;
        }

        }
    }

template <class V>
double CalcMean(std::vector<V> vals, double cutoff)
{
 double sum = 0;
 int cnt = 0;
 for(int i=0; i<vals.size(); i++)
   if(vals[i]>cutoff){ sum += vals[i]; cnt++; }
 return(sum/cnt);

}

int CalcMedian(std::vector<int>& dist)
{
 if(dist.size()==0) return(0.0);
 if(dist.size()==1) return(dist[0]/1.0);

 std::sort(dist.begin(), dist.end());

 if(dist.size()%2==1) return(dist[(dist.size()-1)/2]);

 int index = dist.size()/2;
 int median = int((dist[index-1] + dist[index])/2.0);

 return(median);
}

    // Perform the alalysis associated with this class.  In this case, it is
    // a print of the base & quality information associated with this position.
    virtual void analyze() {
	//if(OUT==NULL) {
	// OUT = fopen(par.outFile.c_str(), "w");
	// if(OUT==NULL) error("Open output file %s failed\n", par.outFile.c_str());
	//}

        if (getRefPosition() != UNSET_POSITION)
        {
	int    t_nt_count[6] = {0,0,0,0,0,0};
	double t_nt_qual[6] = {0,0,0,0,0,0};
	double t_nt_cycle[6] = {0,0,0,0,0,0};
	double t_nt_mapq[6] = {0,0,0,0,0,0};
	int    t_ps_nt_count[6] = {0,0,0,0,0,0};
	double t_ns_nt_count[6] = {0,0,0,0,0,0};
	double t_ps_nt_qual[6] = {0,0,0,0,0,0};
	double t_ns_nt_qual[6] = {0,0,0,0,0,0};
	double t_ps_nt_cycle[6] = {0,0,0,0,0,0};
	double t_ns_nt_cycle[6] = {0,0,0,0,0,0};
	double t_ps_nt_mapq[6] = {0,0,0,0,0,0};
	double t_ns_nt_mapq[6] = {0,0,0,0,0,0};
	int    nt_proper_count[6] = {0,0,0,0,0,0};
	int    ps_nt_proper_count[6] = {0,0,0,0,0,0};
	int    ns_nt_proper_count[6] = {0,0,0,0,0,0};

	int idx = -1;
	char b;
	std::vector<int> mut_dist2ends;
	mut_dist2ends.clear();
	std::vector<int> mut_headClip; // this is not used here. Is this useful?
	mut_headClip.clear();
	double mut_baseq_pass_sum = 0.0;
	double mut_mapq_pass_sum = 0.0;

        for(int i=0; i<this->bases.size(); i++)
	{
		b = toupper(this->bases[i]);

		if(b==mutant)
		{
		 total_mutant_cnt++;
		 if(this->qualities[i]-33 <= par.low_baseq_cutoff)
			mut_low_baseq_cnt++;
		 else mut_baseq_pass_sum += this->qualities[i];

		 if(this->mapq[i] <= par.low_mapq_cutoff)
			mut_low_mapq_cnt++;
		 else mut_mapq_pass_sum += this->mapq[i];

		 if(this->proper[i]==false)
			mut_improper_paired_cnt++;
		 if(this->strands[i]=='+') mut_plus_cnt++;
		 else mut_minus_cnt++;
		 mut_dist2ends.push_back(this->cycles[i]);
		 mut_headClip.push_back(this->headClip[i]);

		}
		else if(b==ref) {
		 total_ref_cnt++;
		 if(this->strands[i]=='+') ref_plus_cnt++;
		 else ref_minus_cnt++;
		}
		else {
		 total_other_allele_cnt++;
		}

		if(this->qualities[i]-33 <= par.low_baseq_cutoff)
			low_baseq_cnt++;
		if(this->mapq[i] <= par.low_mapq_cutoff)
			low_mapq_cnt++;
		if(this->proper[i] == false)
			improper_paired_cnt++;
		total_base_cnt = total_mutant_cnt + total_ref_cnt + total_other_allele_cnt;

		//The following are the average statitics of all bases in the pileup
		if(b=='A') idx = 0; if(b=='C')idx=1; if(b=='G')idx=2; if(b=='T')idx=3; if(b=='N')idx=4;
		if(idx<0) continue;
		nt_count[idx]++;
		if(this->proper[i]==true) nt_proper_count[idx]++;
		t_nt_qual[idx]+=this->qualities[i]-33;
		t_nt_cycle[idx]+=abs(this->cycles[i]);
		t_nt_mapq[idx]+=this->mapq[i];

		if(this->strands[i]=='+')
		{
		ps_nt_count_all[idx]++;
		ps_nt_count[idx]++;
		if(this->proper[i]==true) ps_nt_proper_count[idx]++;
		t_ps_nt_qual[idx]+=this->qualities[i]-33;
		t_ps_nt_cycle[idx]+=abs(this->cycles[i]);
		t_ps_nt_mapq[idx]+=this->mapq[i];
		}
		else if(this->strands[i]=='-')
		{
		ns_nt_count_all[idx]++;
		ns_nt_count[idx]++;
		if(this->proper[i]==true) ns_nt_proper_count[idx]++;
		t_ns_nt_qual[idx]+=this->qualities[i]-33;
		t_ns_nt_cycle[idx]+=abs(this->cycles[i]);
		t_ns_nt_mapq[idx]+=this->mapq[i];
		}
		else error("Strand is not '+' or '-'\n");
		idx = -1;
	}

	for(int i=0; i<5; i++)
	{
		nt_qual[i] = nt_count[i]==0 ? 0 : t_nt_qual[i]/nt_count[i];
		nt_mapq[i] = nt_count[i]==0 ? 0 : t_nt_mapq[i]/nt_count[i];
		nt_cycle[i] = nt_count[i]==0 ? 0 : t_nt_cycle[i]/nt_count[i];
		ps_nt_qual[i] = ps_nt_count[i]==0 ? 0 : t_ps_nt_qual[i]/ps_nt_count[i];
		ps_nt_mapq[i] = ps_nt_count[i]==0 ? 0 : t_ps_nt_mapq[i]/ps_nt_count[i];
		ps_nt_cycle[i] = ps_nt_count[i]==0 ? 0 : t_ps_nt_cycle[i]/ps_nt_count[i];
		ns_nt_qual[i] = ns_nt_count[i]==0 ? 0 : t_ns_nt_qual[i]/ns_nt_count[i];
		ns_nt_mapq[i] = ns_nt_count[i]==0 ? 0 : t_ns_nt_mapq[i]/ns_nt_count[i];
		ns_nt_cycle[i] = ns_nt_count[i]==0 ? 0 : t_ns_nt_cycle[i]/ns_nt_count[i];
	}


	    this->low_mapq_pct = this->total_base_cnt==0 ? 0.0 : double(this->low_mapq_cnt)/this->total_base_cnt * 100.0;
	    this->mut_low_mapq_pct = this->total_mutant_cnt==0 ? 0.0 : double(this->mut_low_mapq_cnt)/this->total_mutant_cnt * 100.0;
	    this->low_baseq_pct = this->total_base_cnt==0 ? 0.0 : double(this->low_baseq_cnt)/this->total_base_cnt * 100.0;
	    this->mut_low_baseq_pct = this->total_mutant_cnt==0 ? 0.0 : double(this->mut_low_baseq_cnt)/this->total_mutant_cnt * 100.0;
	    this->other_allele_pct = double(this->total_other_allele_cnt)/this->total_base_cnt *100.0;
	    this->indel_pct = double(this->insertion_cnt+deletion_cnt)/this->total_base_cnt * 100.0;
	    this->headClip_pct = double(this->headClip_cnt)/this->total_base_cnt * 100.0;

	    mut_median_dist2ends = CalcMedian(mut_dist2ends);

	    double SB_OR, SE, pval;
	    if(mut_minus_cnt*ref_plus_cnt==0 && mut_plus_cnt*ref_minus_cnt==0) SB_OR = -1.0;
	    else SB_OR = double(mut_plus_cnt*ref_minus_cnt)/double(mut_minus_cnt*ref_plus_cnt);
	    if(SB_OR>=0 && SB_OR<1) SB_OR = 1.0/SB_OR;
	    if(mut_minus_cnt>0 && ref_plus_cnt>0 && mut_plus_cnt>0 && ref_minus_cnt)
	    {
		SE = sqrt(1/mut_minus_cnt + 1/ref_plus_cnt + 1/mut_plus_cnt + 1/ref_minus_cnt);
	    }

	    double mut_base_pct = double(total_mutant_cnt)/total_base_cnt*100.0;
	    double mapq_pass_mean = CalcMean(this->mapq, par.low_mapq_cutoff);
	    double baseq_pass_mean = CalcMean(this->qualities, par.low_baseq_cutoff);
	    double mut_mapq_pass_mean = mut_mapq_pass_sum/(total_mutant_cnt-mut_low_mapq_cnt);
	    double mut_baseq_pass_mean = mut_baseq_pass_sum/(total_mutant_cnt-mut_low_baseq_cnt);


	// Construct filter info string
	std::string filter;
	filter="";
	if(insertion_cnt+deletion_cnt >= par.indel_cnt && par.indel_cnt>0 || indel_pct>=par.indel_pct && par.indel_pct>0.0) filter+="INDEL";
	if(mut_median_dist2ends <= par.mut_median_cycles2ends && par.mut_median_cycles2ends>0) filter+= filter.size()==0 ? "CB" : ",CB";
	if(SB_OR >= par.SB_OR) filter += filter.size()==0 ? "SB" :",SB";
	if(total_mutant_cnt<=par.mut_base_cnt && par.mut_base_cnt>0 || mut_base_pct<=par.mut_base_pct && par.mut_base_pct>0.0) filter += filter.size()==0 ? "AB" : ",AB";
	if(headClip_cnt>=par.headClip_cnt && par.headClip_cnt>0 || headClip_pct>=par.headClip_pct && par.headClip_pct>0.0) filter += filter.size()==0 ? "HC" : ",HC";
	if(total_other_allele_cnt>=par.total_other_allele_cnt && par.total_other_allele_cnt>0 || other_allele_pct>=par.other_allele_pct && par.other_allele_pct>0) filter += filter.size()==0 ? "OA" : ",OA";
	if(low_mapq_cnt>=par.low_mapq_cnt && par.low_mapq_cnt>0 || low_mapq_pct>=par.low_mapq_pct && par.low_mapq_pct>0.0) filter += filter.size()==0 ? "MQ" :",MQ";
	if(mut_low_mapq_cnt >= par.mut_low_mapq_cnt && par.mut_low_mapq_cnt>0|| mut_low_mapq_pct >= par.mut_low_mapq_pct && par.mut_low_mapq_pct>0.0) filter += filter.size()==0 ? "MQM" :",MQM";
	if(low_baseq_cnt >= par.low_baseq_cnt && par.low_baseq_cnt>0 || low_baseq_pct >= par.low_baseq_pct && par.low_baseq_pct>0.0) filter += filter.size()==0 ? "BQ" :",BQ";
	if(mut_low_baseq_cnt >= par.mut_low_baseq_cnt && par.mut_low_baseq_cnt>0|| mut_low_baseq_pct >= par.mut_low_baseq_pct && par.mut_low_baseq_pct>0.0) filter += filter.size()==0 ? "BQM" :",BQM";
	if(improper_paired_cnt >= par.improper_paired_cnt && par.improper_paired_cnt>0 ) filter += filter.size()==0 ? "IP" :",IP";
	if(filter.size()==0) filter = ".";

        std::cout << getChromosome() << "\t" << getRefPosition()+1<<"\t"<<getRefPosition()+1<<"\t"<<ref<<"\t"<<mutant<<"\t";
	std::cout << filter << "\t";
        std::cout << total_ref_cnt <<"\t"<< total_mutant_cnt<<"\t"<<total_other_allele_cnt<<"\t"<<ref_plus_cnt<<":"<<ref_minus_cnt<<":"<<mut_plus_cnt<<":"<<mut_minus_cnt<<"\t";
	std::cout << insertion_cnt << "\t" << deletion_cnt << "\t" << headClip_cnt << "\t" << mut_median_dist2ends << "\t";
	if(SB_OR>=0) printf("%.1f\t", SB_OR); else printf(".\t");
	printf("%d\t%d\t", improper_paired_cnt, mut_improper_paired_cnt);
	printf("%d\t%d\t%d\t%d\t", low_mapq_cnt, mut_low_mapq_cnt, low_baseq_cnt, mut_low_baseq_cnt);
	if(total_mutant_cnt-mut_low_baseq_cnt>0) printf("%.1f\t%.1f\t%.1f\t%.1f", mapq_pass_mean, mut_mapq_pass_mean, baseq_pass_mean-33, mut_baseq_pass_mean-33);
	else printf("%.1f\t.\t%.1f\t.", mapq_pass_mean, baseq_pass_mean-33);

/*
	    for(int i=0; i<4; i++) {
	 	std::cout << ps_nt_count_all[i]<< ":" << ns_nt_count_all[i];
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    for(int i=0; i<4; i++) {
	 	std::cout << ps_nt_proper_count[i]<< ":" << ns_nt_proper_count[i];
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";


	    for(int i=0; i<4; i++) {
		printf("%.1f:%.1f", ps_nt_qual[i], ns_nt_qual[i]);
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    for(int i=0; i<4; i++) {
	 	printf("%.1f:%.1f", ps_nt_cycle[i],ns_nt_cycle[i]);
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    for(int i=0; i<4; i++) {
	 	printf("%.1f:%.1f", ps_nt_mapq[i],ns_nt_mapq[i]);
		if(i<3) std::cout << "|";
	    }
	    std::cout<<"\t";

	    if(full)
            for(int i=0; i<this->bases.size(); i++)
            {
            	std::cout << int(this->mapq[i])<<this->bases[i]<< (this->strands[i]=='-' ? 101 - this->cycles[i] : this->cycles[i]) << this->strands[i] <<this->qualities[i]-33<<"\t"; 
            	//std::cout<<cigar_str[i]<<":"<<seq_string[i]<<":"<<qual_string[i]<<"\t";
            }
*/
            std::cout<<"\n";
        }
    };
    void reset(int refPosition) {
        // Call the base class.
        PileupElement::reset(refPosition);
    total_ref_cnt = 0;
    total_mutant_cnt =0;
    total_other_allele_cnt =0;
    insertion_cnt =0;
    deletion_cnt = 0;
    headClip_cnt = 0;
    mut_median_dist2ends = 0;
    low_mapq_cnt = 0;
    low_baseq_cnt = 0;
    mut_low_mapq_cnt = 0;
    mut_low_baseq_cnt = 0;
    improper_paired_cnt = 0;
    mut_improper_paired_cnt = 0;
    low_mapq_pct = 0.0;;
    low_baseq_pct = 0.0;
    mut_low_mapq_pct = 0.0;
    mut_low_baseq_pct = 0.0;
    improper_paired_pct = 0.0;
    mut_improper_paired_pct = 0.0;
    mut_plus_cnt = 0;
    ref_plus_cnt = 0;
    mut_minus_cnt = 0;
    ref_minus_cnt = 0;
	this->alldepth = 0;
        this->depth = 0;
        this->numMinusStrand = 0;
        this->numPlusStrand = 0;
        this->freq.clear();
        this->base.clear();
        this->bases.clear();
        this->qualities.clear();
        this->cycles.clear();
        this->strands.clear();
        this->seq_string.clear();
        this->qual_string.clear();
        this->cigar_str.clear();
	this->mapq.clear();
	this->proper.clear();
	this->headClip.clear();
	for(int i=0;i<6;i++){
 	nt_count[i] = 0; ps_nt_count[i]=0; ns_nt_count[i]=0; ps_nt_count_all[i]=0; ns_nt_count_all[i]=0;
	nt_qual[i] = 0; ps_nt_qual[i]=.0;ns_nt_qual[i]=0;
	nt_cycle[i] = 0; ps_nt_cycle[i]=.0;ns_nt_cycle[i]=0;
	nt_mapq[i] = 0; ps_nt_mapq[i]=.0;ns_nt_mapq[i]=0;
	}
	r1f=r1r=r2f=r2r=0;
    };

public:
    U depth;
    T freq; // freq of A, C, G, T
    CmdParameters par;
    char ref;
    char mutant;
    bool parset;
    int total_base_cnt;
    int total_ref_cnt;
    int total_mutant_cnt;
    int total_other_allele_cnt;
    int other_allele_pct;
    int insertion_cnt;
    int deletion_cnt;
    double indel_pct;
    int headClip_cnt;
    double headClip_pct;
    int mut_median_dist2ends;
    int low_mapq_cnt;
    int low_baseq_cnt;
    int mut_low_mapq_cnt;
    int mut_low_baseq_cnt;
    int improper_paired_cnt;
    int mut_improper_paired_cnt;
    double low_mapq_pct;
    double low_baseq_pct;
    double mut_low_mapq_pct;
    double mut_low_baseq_pct;
    double improper_paired_pct;
    double mut_improper_paired_pct;
    int mut_plus_cnt;
    int ref_plus_cnt;
    int mut_minus_cnt;
    int ref_minus_cnt;

    int alldepth;
    int nt_count[6];
    int ps_nt_count[6];
    int ns_nt_count[6];
    int ps_nt_count_all[6];
    int ns_nt_count_all[6];
    double nt_qual[6];
    double ps_nt_qual[6];
    double ns_nt_qual[6];
    double nt_cycle[6];
    double ps_nt_cycle[6];
    double ns_nt_cycle[6];
    double nt_mapq[6];
    double ps_nt_mapq[6];
    double ns_nt_mapq[6];
    int numPlusStrand;
    int numMinusStrand;
    std::string base;
    std::vector<char> mapq;
    std::vector<char> bases;
    std::vector<int> qualities;
    std::vector<int> cycles;
    std::vector<char> strands;
    std::vector<std::string> seq_string;
    std::vector<std::string> qual_string;
    std::vector<unsigned int> mapPos;
    std::vector<std::string> cigar_str;
    std::vector<bool> proper;
    std::vector<int> headClip;
    int r1f, r1r, r2f, r2r;
public:
    FILE *OUT;
};

template <class T>
class RegionPileup: public Pileup<T>{

public:
  CmdParameters par;
  std::vector<BedInfo> sites;
  std::map<char, int> allele2Int;
  char ref;
  char mutant;
  // Let's see if we can manage these in a more economical fashion. EST - March 15, 2013
  SamFile samIn;
  SamFileHeader header;

public:
 void InitAllele2IntMap()
 {
  allele2Int['A'] = 0;
  allele2Int['C'] = 1;
  allele2Int['G'] = 2;
  allele2Int['T'] = 3;
 }

  RegionPileup():Pileup<T>() { }
  RegionPileup(int window):Pileup<T>(window) { };
	RegionPileup(int window, char *bamFile) : Pileup<T>(window) {}

	/**
	  * Here, you will initialize your bam, including an accompanying index file. 
	  * This should be done before doing any other stuff like setting a position
	  */

 void PrintHeader()
 {
  printf("#CHR\tSTART\tEND\tREF\tMUT\tFILTER\tnRef\tnMut\tnOA\tRF:RR:MF:MR\tnINS\tnDEL\tnClip_mut\tCycle_mut\tSB\tnIP\tnIP_mut\tnLMQ\tnLMQ_mut\tnLBQ\tnLBQ_mut\tmeanMQ_pass\tmut_meanMQ_pass\tmeanBQ_pass\tmut_meanBQ_pass\n");
 }

 void LoadBedFile(std::string bedfile)
 {
  FILE *bed = fopen(bedfile.c_str(), "r");
  if(bed==NULL) error("Open BED file failed\n");

  String buffer;
  StringArray tokens;
  BedInfo bedinfo;

  while(!feof(bed))
    {
      buffer.ReadLine(bed);
      tokens.ReplaceTokens(buffer);
      if(tokens.Length()==0) continue;
      if(tokens.Length()<5) error("BED file should have at leat 5 columns: %s\n", buffer.c_str() );
      bedinfo.chr = std::string(tokens[0].c_str());
      bedinfo.start = tokens[1].AsInteger();
      bedinfo.end = tokens[2].AsInteger();
      if(bedinfo.start != bedinfo.end) error("Start should be equal to end: %s\n", buffer.c_str() );
      bedinfo.ref = tokens[3][0];
      bedinfo.mutant = tokens[4][0];
      sites.push_back(bedinfo);
    }
  fclose(bed);

 }

 int  init(const char *bamfile) {
	// We first attemp to open our bam file
	if(!samIn.OpenForRead(bamfile)) {
		fprintf(stderr, "%s\n", samIn.GetStatusMessage());
		return(samIn.GetStatus());
	}
	if(!samIn.ReadHeader(header))
	{
		fprintf(stderr, "%s\n", samIn.GetStatusMessage());
		return(samIn.GetStatus());
	}
	std::string _indexFileName = std::string(bamfile) + ".bai";

	// Next, we need to open up the index file
	if (!samIn.ReadBamIndex(_indexFileName.c_str())) {
		fprintf(stderr, "%s\n", samIn.GetStatusMessage());
		return(samIn.GetStatus());
	};
	return SamStatus::SUCCESS;
 }


 void addElement(T& element, SamRecord& record)
 {
    if(element.parset==false)
    {
      element.par = par; element.parset = true;
    }
    element.ref = ref;
    element.mutant = mutant;
    element.addEntry(record);
 }

 virtual int processFile(const std::string& fileName, 
                         uint16_t excludeFlag = 0x0704, 
                         uint16_t includeFlag = 0) {
        SamFile samIn;
        SamFileHeader header;
        SamRecord record;

        if(!samIn.OpenForRead(fileName.c_str()))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }

        // The file needs to be sorted by coordinate.
        samIn.setSortedValidation(SamFile::COORDINATE);

        // Iterate over all records
        while (samIn.ReadRecord(header, record))
        {
            uint16_t flag = record.getFlag();
            if(flag & 0x4 || flag & 0x200) continue; //unmapped or failed
	    if(flag & 0x400) continue; //duplicate

            if(flag & excludeFlag)
            {
                // This record has an excluded flag set, 
                // so continue to the next one.
                continue;
            }
            if((flag & includeFlag) != includeFlag)
            {
                // This record does not have all required flags set, 
                // so continue to the next one.
                continue;
            }
            Pileup<T>::processAlignment(record);
        }

        Pileup<T>::flushPileup();

        int returnValue = 0;
        if(samIn.GetStatus() != SamStatus::NO_MORE_RECS)
        {
            // Failed to read a record.
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            returnValue = samIn.GetStatus();
        }
        return(returnValue);  
    };

    int processRegion(const char*  refName,
                      int start,
                      int end) {

        SamRecord record;
        if (!samIn.SetReadSection(refName, start, end))
        {
            fprintf(stderr, "%s\n", samIn.GetStatusMessage());
            return(samIn.GetStatus());
        }

        int numSectionRecords;
        // Set returnStatus to success.  It will be changed
        // to the failure reason if any of the writes fail.
        SamStatus::Status returnStatus = SamStatus::SUCCESS;

        while(samIn.ReadRecord(header, record))
        {
					uint16_t flag = record.getFlag();
					if(flag & 0x4 || flag&0x200) continue; //unmaped or failed
					if(flag & 0x400) continue; //duplicate

					// Successfully read a record from the file, so pile it up.
					this->processAlignment(record, start, end);
					++numSectionRecords;
        }
			Pileup<T>::flushPileup();

			return (returnStatus);
    };

    virtual void processAlignment(SamRecord& record, 
                                  const int begin, 
                                  const int end) {
        int refPosition = record.get0BasedPosition();
        int refID = record.getReferenceID();

        // Flush any elements from the pileup that are prior to this record
        // since the file is sorted, we are done with those positions.
        Pileup<T>::flushPileup(refID, refPosition);

        // Loop through for each reference position covered by the record.
        // It is up to the PILEUP_TYPE to handle insertions/deletions, etc
        // that are related with the given reference position.

    	if(end-begin>1)
         for(; refPosition <= record.get0BasedAlignmentEnd(); ++refPosition)
         {
	    //Here begin is 0-based and end is 1-based. Confusing...
            // check refPosition is in specified region
            if (begin <= refPosition && refPosition < end)
              Pileup<T>::addAlignmentPosition(refPosition, record);

         }
	else // for single point it directly uses the begin as the reference position
	  Pileup<T>::addAlignmentPosition(begin, record);
    }
};

#define PILEUP_WINDOW_SIZE 1024

int main(int argc, char *argv[])
{
String sites;
String outFile;
int min_mapq = 0;
int min_baseq = 0;
int minDepth = 0;
int maxDepth = -1;

StringArray bamFiles;
String bam;

int indel_winsize = 10;
int indel_cnt = 3;
double indel_pct = -1.0;
int mut_median_cycles2ends = 5;
int mut_base_cnt = -1;
double mut_base_pct = 30;
int headClip_cnt = -1;
double headClip_pct = -1.0;
int zero_mapq = 0;
int improper_paired_cnt = -1;
int mut_improper_paired_cnt = -1;
double improper_paired_pct = -1.0;
double mut_improper_paired_pct = 10;
double low_mapq_cutoff = 10;
int low_mapq_cnt = -1;
double low_mapq_pct = 10;
int mut_low_mapq_cnt = -1;
double mut_low_mapq_pct = -1.0;
double low_baseq_cutoff = 20;
int low_baseq_cnt = -1;
double low_baseq_pct = 10;
int mut_low_baseq_cnt = -1;
double mut_low_baseq_pct = -1.0;
double strand_bias_OR = 2;
int total_other_allele_cnt = -1;
double other_allele_pct = 10;

  ParameterList pl;

  BEGIN_LONG_PARAMETERS(longParameters)
    LONG_PARAMETER_GROUP("Input Files")
      LONG_STRINGPARAMETER("bam",&bam)
      LONG_STRINGPARAMETER("bed",&sites)
    //LONG_PARAMETER_GROUP("Output File")
    //  LONG_STRINGPARAMETER("out",&outFile)
   // LONG_PARAMETER_GROUP("Exclusion Criteria")
   //   LONG_INTPARAMETER("min_mapq",&min_mapq)
   //   LONG_INTPARAMETER("min_baseq",&min_baseq)
    LONG_PARAMETER_GROUP("Cycle Bias (CB)")
      LONG_INTPARAMETER("mut_median_cycles2ends",&mut_median_cycles2ends)
    LONG_PARAMETER_GROUP("Strand Bias (SB)")
      LONG_DOUBLEPARAMETER("SB_OR",&strand_bias_OR)
    LONG_PARAMETER_GROUP("Nearby Indels (INDEL)")
      LONG_INTPARAMETER("indel_winsize",&indel_winsize)
      LONG_INTPARAMETER("indel_cnt",&indel_cnt)
      LONG_DOUBLEPARAMETER("indel_pct",&indel_pct)
    LONG_PARAMETER_GROUP("Read Clipping (CL)")
      LONG_INTPARAMETER("mut_clip_cnt",&headClip_cnt)
      LONG_DOUBLEPARAMETER("mut_clip_pct",&headClip_pct)
    LONG_PARAMETER_GROUP("Other Alleles (OA)")
      LONG_INTPARAMETER("other_allele_cnt",&total_other_allele_cnt)
      LONG_DOUBLEPARAMETER("other_allele_pct",&other_allele_pct)
    LONG_PARAMETER_GROUP("Allelic Balance (AB)")
      LONG_INTPARAMETER("mut_base_cnt",&mut_base_cnt)
      LONG_DOUBLEPARAMETER("mut_base_pct",&mut_base_pct)
    LONG_PARAMETER_GROUP("Low Map Quality (LMQ)")
      LONG_DOUBLEPARAMETER("low_mapq_cutoff",&low_mapq_cutoff)
      LONG_INTPARAMETER("low_mapq_cnt",&low_mapq_cnt)
      LONG_DOUBLEPARAMETER("low_mapq_pct",&low_mapq_pct)
      LONG_INTPARAMETER("mut_low_mapq_cnt",&mut_low_mapq_cnt)
      LONG_DOUBLEPARAMETER("mut_low_mapq_pct",&mut_low_mapq_pct)
    LONG_PARAMETER_GROUP("Low Base Quality (LBQ)")
      LONG_DOUBLEPARAMETER("low_baseq_cutoff",&low_baseq_cutoff)
      LONG_INTPARAMETER("low_baseq_cnt",&low_baseq_cnt)
      LONG_DOUBLEPARAMETER("low_baseq_pct",&low_baseq_pct)
      LONG_INTPARAMETER("mut_low_baseq_cnt",&mut_low_baseq_cnt)
      LONG_DOUBLEPARAMETER("mut_low_baseq_pct",&mut_low_baseq_pct)
    LONG_PARAMETER_GROUP("Improper Paired (IP)")
      LONG_INTPARAMETER("improper_paried_cnt",&improper_paired_cnt)
      LONG_DOUBLEPARAMETER("improper_paried_pct",&improper_paired_pct)
      LONG_INTPARAMETER("mut_improper_paried_cnt",&mut_improper_paired_cnt)
      LONG_DOUBLEPARAMETER("mut_improper_paried_pct",&mut_improper_paired_pct)
   END_LONG_PARAMETERS();

  pl.Add(new LongParameters("", longParameters));
  pl.Read(argc, argv);

  pl.Status();

  fprintf(stderr, "NOTE: When parameters are negative these filters are NOT in effect!\n");
  fprintf(stderr, "      When filters have 'mut_' the filters are for the statistics calculated for MUTANT alleles only!\n\n");

  CmdParameters par;

  par.min_baseq = min_baseq;
  par.min_mapq = min_mapq;

  par.indel_winsize = indel_winsize;
  par.indel_cnt = indel_cnt;
  par.indel_pct = indel_pct;
  par.headClip_cnt = headClip_cnt;
  par.headClip_pct = headClip_pct;
  par.mut_median_cycles2ends = mut_median_cycles2ends;
  par.mut_base_cnt = mut_base_cnt;
  par.mut_base_pct = mut_base_pct;
  par.zero_mapq = zero_mapq;;
  par.improper_paired_cnt = improper_paired_cnt;
  par.mut_improper_paired_cnt = mut_improper_paired_cnt;
  par.low_mapq_cutoff = low_mapq_cutoff;
  par.low_mapq_cnt = low_mapq_cnt;
  par.low_mapq_pct = low_mapq_pct;
  par.mut_low_mapq_cnt = mut_low_mapq_cnt;
  par.mut_low_mapq_pct = mut_low_mapq_pct;
  par.low_baseq_cutoff = low_baseq_cutoff;
  par.low_baseq_cnt = low_baseq_cnt;
  par.low_baseq_pct = low_baseq_pct;
  par.mut_low_baseq_cnt = mut_low_baseq_cnt;
  par.mut_low_baseq_pct = mut_low_baseq_pct;
  par.strand_bias_OR = strand_bias_OR;
  par.SB_OR = strand_bias_OR;
  par.total_other_allele_cnt = total_other_allele_cnt;
  par.other_allele_pct = other_allele_pct;
  par.outFile = std::string(outFile.c_str());

  if(bam.Length()==0) error("No bam file provided!\n");
  //if(outFile.Length()==0) error("No output file provided!\n");

    //printf("processing %s...\n", argv[1]);
    if (sites.Length()>0) {
        std::string chr;
	std::string ref;
	std::string mutant;
        int begin;
        int end;

	RegionPileup<SiteCoverage<freqType, int > > pileup(PILEUP_WINDOW_SIZE);
	pileup.init(bam.c_str());
	pileup.InitAllele2IntMap();
	pileup.par = par;

	pileup.LoadBedFile(std::string(sites.c_str()));
	pileup.PrintHeader();

        for (unsigned int i = 0; i < pileup.sites.size(); i++){
            chr = pileup.sites[i].chr;
            begin = pileup.sites[i].start;
            end = pileup.sites[i].end;
            pileup.ref = pileup.sites[i].ref;
            pileup.mutant = pileup.sites[i].mutant;

	    if(chr.compare("")==0 || begin<=0 || end<begin || begin!=end) error("Sites are not correctly specified: chr=%s begin=%d end=%d\n", chr.c_str(), begin, end);

            pileup.processRegion(chr.c_str(), begin-1, end);
        }
    }
    if (sites.Length()==0) {
        Pileup<SiteCoverage<freqType, int > > pileup(PILEUP_WINDOW_SIZE);
        pileup.processFile(std::string(bam.c_str()));
    }
    return 0;
}
